@@ -189,7 +189,9 @@ extern char *var_relayhost;
 #define DEF_SND_RELAY_MAPS	""
 extern char *var_snd_relay_maps;
 
-#define VAR_FALLBACK_RELAY	"fallback_relay"
+#define VAR_SMTP_FALLBACK	"smtp_fallback_relay"
+#define DEF_SMTP_FALLBACK	"$fallback_relay"
+#define VAR_LMTP_FALLBACK	"smtp_fallback_relay"
 #define DEF_FALLBACK_RELAY	""
 extern char *var_fallback_relay;
 
