@@ -116,7 +116,6 @@ static SMTP_SESSION *smtp_connect_sock(int, struct sockaddr *, int,
 /* smtp_connect_unix - connect to UNIX-domain address */
 
 static SMTP_SESSION *smtp_connect_unix(const char *addr,
-				               const char *destination,
 				               DSN_BUF *why,
 				               int sess_flags)
 {
@@ -163,7 +162,7 @@ static SMTP_SESSION *smtp_connect_unix(const char *addr,
 	msg_info("%s: trying: %s...", myname, addr);
 
     return (smtp_connect_sock(sock, (struct sockaddr *) & sun, sizeof(sun),
-			      addr, addr, 0, destination, why, sess_flags));
+			      addr, addr, 0, addr, why, sess_flags));
 }
 
 /* smtp_connect_addr - connect to explicit address */
@@ -362,7 +361,7 @@ static SMTP_SESSION *smtp_connect_sock(int sock, struct sockaddr * sa,
 static char *smtp_parse_destination(char *destination, char *def_service,
 				            char **hostp, unsigned *portp)
 {
-    char   *buf;
+    char   *buf = mystrdup(destination);
     char   *service;
     struct servent *sp;
     char   *protocol = "tcp";		/* XXX configurable? */
@@ -372,27 +371,6 @@ static char *smtp_parse_destination(char *destination, char *def_service,
     if (msg_verbose)
 	msg_info("smtp_parse_destination: %s %s", destination, def_service);
 
-    /*
-     * XXX Support for UNIX-domain destinations requires that we split the
-     * smtp_connect() function into the portions that are dedicated to TCP,
-     * and the portions that can be shared with TCP and UNIX-domain sockets.
-     */
-    if (strncmp(destination, "unix:", 5) == 0) {
-	msg_fatal("UNIX-domain is currently unimplemented");
-	*hostp = buf = mystrdup(destination + 5);
-	*portp = 0;
-	return (buf);
-    }
-
-    /*
-     * XXX And while we're fixing that, we should also ensure that the inet:
-     * prefix is recognized with LMTP destinations only.
-     */
-    if (strncmp(destination, "inet:", 5) == 0)
-	buf = mystrdup(destination + 5);
-    else
-	buf = mystrdup(destination);
-
     /*
      * Parse the host/port information. We're working with a copy of the
      * destination argument so the parsing can be destructive.
@@ -490,6 +468,58 @@ static void smtp_cleanup_session(SMTP_STATE *state)
     request->msg_stats.reuse_count = 0;
 }
 
+/* smtp_connect_local - connect to local server */
+
+static void smtp_connect_local(SMTP_STATE *state, const char *path,
+			               DSN_BUF *why)
+{
+    DELIVER_REQUEST *request = state->request;
+    int     sess_flags = SMTP_SESS_FLAG_NONE;
+    SMTP_SESSION *session;
+
+    /*
+     * Connection cache management is based on the UNIX-domain pathname,
+     * without the "unix:" prefix.
+     * 
+     * XXX Disable connection caching when sender-dependent authentication is
+     * enabled. We must not send someone elses mail over an authenticated
+     * connection, and we must not send mail that requires authentication
+     * over a connection that wasn't authenticated.
+     */
+#define CAN_ENABLE_CONN_CACHE(request, dest) \
+    (!var_smtp_sender_auth \
+     && ((var_smtp_cache_demand && (request->flags & DEL_REQ_FLAG_SCACHE)) \
+	 || (smtp_cache_dest && string_list_match(smtp_cache_dest, dest))))
+
+    if (CAN_ENABLE_CONN_CACHE(request, path))
+	sess_flags |= SMTP_SESS_FLAG_CACHE;
+
+    /*
+     * XXX We assume that the session->addr member refers to a copy of the
+     * UNIX-domain pathname, so that smtp_save_session() will cache the
+     * connection using the pathname as the physical endpoint name.
+     */
+#define NO_MX	0
+#define NO_PORT	0
+
+    if ((sess_flags & SMTP_SESS_FLAG_CACHE) == 0
+	|| (session = smtp_reuse_addr(state, path, NO_PORT)) == 0)
+	session = smtp_connect_unix(path, why, sess_flags);
+    if ((state->session = session) != 0) {
+	session->state = state;
+	state->final_server = 1;
+	if ((session->features & SMTP_FEATURE_FROM_CACHE) == 0
+	    && smtp_helo(state) != 0) {
+	    if (vstream_ferror(session->stream) == 0
+		&& vstream_feof(session->stream) == 0)
+		smtp_quit(state);
+	} else {
+	    smtp_xfer(state);
+	}
+	smtp_cleanup_session(state);
+    }
+}
+
 /* smtp_scrub_address_list - delete all cached addresses from list */
 
 static void smtp_scrub_addr_list(HTABLE *cached_addr, DNS_RR **addr_list)
@@ -574,6 +604,7 @@ static int smtp_reuse_session(SMTP_STATE *state, int lookup_mx,
     DNS_RR *addr;
     DNS_RR *next;
     int     saved_final_server = state->final_server;
+    MAI_HOSTADDR_STR hostaddr;
     SMTP_SESSION *session;
 
     /*
@@ -603,7 +634,8 @@ static int smtp_reuse_session(SMTP_STATE *state, int lookup_mx,
 	if (addr->pref != domain_best_pref)
 	    break;
 	next = addr->next;
-	if ((session = smtp_reuse_addr(state, addr, port)) != 0) {
+	if (dns_rr_to_pa(addr, &hostaddr) != 0
+	    && (session = smtp_reuse_addr(state, hostaddr.buf, port)) != 0) {
 	    session->features |= SMTP_FEATURE_BEST_MX;
 	    session_count += 1;
 	    smtp_update_addr_list(addr_list, session->addr, session_count);
@@ -617,12 +649,12 @@ static int smtp_reuse_session(SMTP_STATE *state, int lookup_mx,
     return (session_count);
 }
 
-/* smtp_connect - establish SMTP connection */
+/* smtp_connect_remote - establish remote connection */
 
-int     smtp_connect(SMTP_STATE *state)
+static void smtp_connect_remote(SMTP_STATE *state, const char *nexthop,
+				        DSN_BUF *why)
 {
     DELIVER_REQUEST *request = state->request;
-    DSN_BUF *why = dsb_create();
     char   *dest_buf;
     char   *domain;
     unsigned port;
@@ -641,6 +673,7 @@ int     smtp_connect(SMTP_STATE *state)
     int     sess_flags = SMTP_SESS_FLAG_NONE;
     int     i_am_mx = 0;
     int     non_fallback_sites;
+    MAI_HOSTADDR_STR hostaddr;
 
     def_service = (state->misc_flags & SMTP_MISC_FLAG_USE_LMTP) ?
 	var_lmtp_tcp_port : "smtp";		/* XXX ##IPPORT_SMTP? */
@@ -657,7 +690,8 @@ int     smtp_connect(SMTP_STATE *state)
     if (sites->argc == 0)
 	msg_panic("null destination: \"%s\"", request->nexthop);
     non_fallback_sites = sites->argc;
-    argv_split_append(sites, var_fallback_relay, ", \t\r\n");
+    if ((state->misc_flags & SMTP_MISC_FLAG_USE_LMTP) == 0)
+	argv_split_append(sites, var_fallback_relay, ", \t\r\n");
 
     /*
      * Don't give up after a hard host lookup error until we have tried the
@@ -735,10 +769,7 @@ int     smtp_connect(SMTP_STATE *state)
 	 * authenticated connection, and we must not send mail that requires
 	 * authentication over a connection that wasn't authenticated.
 	 */
-	if (cpp == sites->argv
-	    && !var_smtp_sender_auth
-	    && ((var_smtp_cache_demand && (request->flags & DEL_REQ_FLAG_SCACHE) != 0)
-		|| (smtp_cache_dest && string_list_match(smtp_cache_dest, domain)))) {
+	if (cpp == sites->argv && CAN_ENABLE_CONN_CACHE(request, domain)) {
 	    sess_flags |= SMTP_SESS_FLAG_CACHE;
 	    SET_NEXTHOP_STATE(state, lookup_mx, domain, port);
 	}
@@ -800,7 +831,8 @@ int     smtp_connect(SMTP_STATE *state)
 		next = 0;
 	    if ((sess_flags & SMTP_SESS_FLAG_CACHE) == 0
 		|| addr->pref == domain_best_pref
-		|| (session = smtp_reuse_addr(state, addr, port)) == 0)
+		|| dns_rr_to_pa(addr, &hostaddr) == 0
+		|| !(session = smtp_reuse_addr(state, hostaddr.buf, port)))
 		session = smtp_connect_addr(dest, addr, port, why, sess_flags);
 	    if ((state->session = session) != 0) {
 		session->state = state;
@@ -832,35 +864,35 @@ int     smtp_connect(SMTP_STATE *state)
     /*
      * We still need to deliver, bounce or defer some left-over recipients:
      * either mail loops or some backup mail server was unavailable.
-     * 
-     * Pay attention to what could be configuration problems, and pretend that
-     * these are recoverable rather than bouncing the mail.
-     * 
-     * In case of a "no error" indication we make up an excuse; this can happen
-     * when the fall-back relay was already tried via a cached connection, so
-     * that the address list scrubber left behind an empty list.
      */
     if (SMTP_RCPT_LEFT(state) > 0) {
+
+	/*
+	 * In case of a "no error" indication we make up an excuse; this can
+	 * happen when the fall-back relay was already tried via a cached
+	 * connection, so that the address list scrubber left behind an empty
+	 * list.
+	 */
 	if (smtp_errno == SMTP_ERR_NONE) {
 	    smtp_dsn_update(why, DSN_BY_LOCAL_MTA,
 			    "4.3.0", 450, "450 Server unavailable",
 			    "server unavailable or unable to receive mail");
 	    smtp_errno = SMTP_ERR_RETRY;
 	}
-	switch (smtp_errno) {
 
-	default:
-	    msg_panic("smtp_connect: bad error indication %d", smtp_errno);
-
-	case SMTP_ERR_LOOP:
-	case SMTP_ERR_FAIL:
+	/*
+	 * Pay attention to what could be configuration problems, and pretend
+	 * that these are recoverable rather than bouncing the mail.
+	 */
+	else if (smtp_errno != SMTP_ERR_RETRY
+		 && (state->misc_flags & SMTP_MISC_FLAG_USE_LMTP) == 0) {
 
 	    /*
 	     * The fall-back destination did not resolve as expected, or it
 	     * is refusing to talk to us, or mail for it loops back to us.
 	     */
 	    if (IS_FALLBACK_RELAY(cpp, sites, non_fallback_sites)) {
-		msg_warn("%s configuration problem", VAR_FALLBACK_RELAY);
+		msg_warn("%s configuration problem", VAR_SMTP_FALLBACK);
 		vstring_strcpy(why->status, "4.3.5");
 		/* XXX Keep the diagnostic code and MTA. */
 		smtp_errno = SMTP_ERR_RETRY;
@@ -885,36 +917,9 @@ int     smtp_connect(SMTP_STATE *state)
 		state->status = deliver_pass_all(MAIL_CLASS_PRIVATE,
 						 var_bestmx_transp,
 						 request);
+		smtp_errno = 0;			/* XXX */
 		SMTP_RCPT_LEFT(state) = 0;	/* XXX */
-		break;
 	    }
-	    /* FALLTHROUGH */
-
-	case SMTP_ERR_RETRY:
-
-	    /*
-	     * We still need to bounce or defer some left-over recipients:
-	     * either mail loops or some mail server was unavailable.
-	     * 
-	     * XXX Unlike enhanced status codes, changing a 4xx into 5xx SMTP
-	     * code is not simply a matter of changing the initial digit.
-	     * What we're doing here is correct only under specific
-	     * conditions, such as changing 450 into 550 or vice versa.
-	     */
-	    state->final_server = 1;		/* XXX */
-	    if (smtp_errno == SMTP_ERR_RETRY)
-		STR(why->status)[0] = STR(why->dtext)[0] = '4';	/* XXX */
-	    else
-		STR(why->status)[0] = STR(why->dtext)[0] = '5';	/* XXX */
-	    why->dcode = atoi(STR(why->dtext));	/* XXX */
-	    smtp_sess_fail(state, why);
-
-	    /*
-	     * Sanity check. Don't silently lose recipients.
-	     */
-	    smtp_rcpt_cleanup(state);
-	    if (SMTP_RCPT_LEFT(state) > 0)
-		msg_panic("smtp_connect: left-over recipients");
 	}
     }
 
@@ -924,6 +929,80 @@ int     smtp_connect(SMTP_STATE *state)
     if (HAVE_NEXTHOP_STATE(state))
 	FREE_NEXTHOP_STATE(state);
     argv_free(sites);
+}
+
+/* smtp_connect - establish SMTP connection */
+
+int     smtp_connect(SMTP_STATE *state)
+{
+    DELIVER_REQUEST *request = state->request;
+    DSN_BUF *why = dsb_create();
+    char   *destination = request->nexthop;
+
+    /*
+     * All deliveries proceed along the same lines, whether they are over TCP
+     * or UNIX-domain sockets, and whether they use SMTP or LMTP: get a
+     * connection from the cache or create a new connection; deliver mail;
+     * update the connection cache or disconnect.
+     * 
+     * The major differences appear at a higher level: the expansion from
+     * destination to address list, and whether to stop before we reach the
+     * end of that list.
+     */
+
+    /*
+     * With LMTP we have direct-to-host delivery only. The destination may
+     * have multiple IP addresses, but that is where the complexity ends. The
+     * functions that do the work mostly are ignorant of whether we use SMTP
+     * or LMTP.
+     */
+    if (state->misc_flags & SMTP_MISC_FLAG_USE_LMTP) {
+	if (strncmp(destination, "unix:", 5) == 0) {
+	    smtp_connect_local(state, destination + 5, why);
+	} else {
+	    if (strncmp(destination, "inet:", 5) == 0)
+		destination += 5;
+	    smtp_connect_remote(state, destination, why);
+	}
+    }
+
+    /*
+     * With SMTP we can have indirection via MX host lookup, as well as an
+     * optional fall-back relayhost.
+     * 
+     * XXX We don't add support for "unix:" or "inet:" prefixes in SMTP
+     * destinations, because that would break compatibility with existing
+     * Postfix configurations that have a host with such a name.
+     */
+    else {
+	smtp_connect_remote(state, destination, why);
+    }
+
+    /*
+     * We still need to bounce or defer some left-over recipients: either
+     * (SMTP) mail loops or some mail server was unavailable.
+     * 
+     * XXX Unlike enhanced status codes, changing a 4xx into 5xx SMTP code is
+     * not simply a matter of changing the initial digit. What we're doing
+     * here is correct only under specific conditions, such as changing 450
+     * into 550 or vice versa.
+     */
+    if (SMTP_RCPT_LEFT(state) > 0) {
+	state->final_server = 1;		/* XXX */
+	if (smtp_errno == SMTP_ERR_RETRY)
+	    STR(why->status)[0] = STR(why->dtext)[0] = '4';	/* XXX */
+	else
+	    STR(why->status)[0] = STR(why->dtext)[0] = '5';	/* XXX */
+	why->dcode = atoi(STR(why->dtext));	/* XXX */
+	smtp_sess_fail(state, why);
+
+	/*
+	 * Sanity check. Don't silently lose recipients.
+	 */
+	smtp_rcpt_cleanup(state);
+	if (SMTP_RCPT_LEFT(state) > 0)
+	    msg_panic("smtp_connect: left-over recipients");
+    }
     dsb_free(why);
     return (state->status);
 }
